/*==========================================================================
 * Copyright (c) 2014, Bisnode Credit AS, All Rights Reserved.
 *
 * This software is the confidential and proprietary information
 * of Bisnode Credit AS ("Confidential Information"). You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Bisnode Credit AS.
 *
 * @since 04.july 2008
=============================================================================*/
package no.soliditet.wsclients.ws3.jaxrpc.client;

import java.io.IOException;
import java.net.MalformedURLException;
import java.rmi.RemoteException;

import javax.xml.rpc.ServiceException;

import no.soliditet.core.exceptionhandling.ExternalInterfacesXmlException;
import no.soliditet.ptpTest.jaxrpc.client.InformationRequestXMLReader;
import no.soliditet.ptpTest.jaxrpc.client.ScoringRequestXMLReader;
import no.soliditet.ptpTest.jaxrpc.client.WSRequestAndResponseClientWriter;
import no.soliditet.wsclients.RequestInfo;
import no.soliditet.wsclients.ws3.jaxrpc.generated.ArrayOfDmWSApplicant;
import no.soliditet.wsclients.ws3.jaxrpc.generated.ArrayOfDmWSElement;
import no.soliditet.wsclients.ws3.jaxrpc.generated.ArrayOfDmWSSingleRequest;
import no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSApplicant;
import no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSElement;
import no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSSingleRequest;

public class DecisionMakerWebServices3Client {

   public static String getVersion(String protocol, String hostname) throws MalformedURLException, ServiceException, RemoteException {
      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSVersion version = getWs(protocol, hostname).getVersion();
      return "DM: " + version.getDmVersion() + ". WS: " + version.getWsVersion();
   }

   public static String getStatus(String protocol, String hostname, String user, String pwd) throws MalformedURLException, ServiceException,
         RemoteException {
      String str = "";
      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSStatus[] s = getWs(protocol, hostname).getStatus(user, pwd);
      if ((s != null) && (s.length > 0)) {
         for (no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSStatus ws : s) {
            if (ws != null) {
               str += ws.getName() + ": " + ws.getStatus() + "\r\n";
            }
         }
      }
      return str;
   }

   /**
    * Makes a scoring using the DecisionaMaker Web Services API. Batch requests are not supported.
    * 
    * @param protocol http or https
    * @param hostname hostname and portnumber (e.g., myhost:8080)
    * @param requestXML the input encoded as dmRequest as specified for PTP where the appInfo (if present) is not encoded in '![CDATA[<appInfo>]]'
    *           element.
    * @return the score returned from DecisionMaker encoded as XML
    * @throws IOException on error
    * @throws ServiceException on error
    * @throws ExternalInterfacesXmlException on error
    */
   public static String getScoreFromXML(String protocol, String hostname, String requestXML) throws IOException, ServiceException,
         ExternalInterfacesXmlException {
      // Convert request XML to a WS request
      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSRequest request = new InformationRequestXMLReader().fromXML(requestXML);

      ArrayOfDmWSSingleRequest arrayOfrqs = request.getRequests();

      if ((arrayOfrqs != null) && (arrayOfrqs.getItem() != null) && (arrayOfrqs.getItem().length > 0)) {
         DmWSSingleRequest[] rqs = arrayOfrqs.getItem();
         RequestInfo info = getInfoFromRequest(rqs);
         no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSInformationRequest sr = (no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSInformationRequest)rqs[0];
         String requestId = sr.getRequestId();
         boolean extended = false;
         int pos = sr.getInformationHandler().indexOf("extended");
         if (pos >= 0) {
            extended = true;
         }

         DecisionMakerWebServices3Impl ws = getWs(protocol, hostname);
         no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSResponse response = ws.getScoreInformationById(request.getUser(), request.getPassWord(),
               requestId, info.scoreId, extended);
         // Converting result back to XML
         return new WSRequestAndResponseClientWriter().responseToXML(response, rqs.length);
      }
      else
         return null;
   }

   private static no.soliditet.wsclients.ws3.jaxrpc.client.DecisionMakerWebServices3Impl getWs(String protocol, String hostname)
         throws MalformedURLException, ServiceException {
      if ((protocol == null) || (protocol.length() == 0)) {
         protocol = "http";
      }
      String endPointURL = protocol + "://" + hostname;
      return new DecisionMakerWebServices3Impl(endPointURL);
   }

   public static String updateScoreFromXML(String protocol, String hostname, String requestXML) throws IOException, ServiceException,
         ExternalInterfacesXmlException {
      // Convert request XML to a WS request
      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSRequest request = new InformationRequestXMLReader().fromXML(requestXML);

      ArrayOfDmWSSingleRequest arrayOfrqs = request.getRequests();

      if ((arrayOfrqs != null) && (arrayOfrqs.getItem() != null) && (arrayOfrqs.getItem().length > 0)) {
         DmWSSingleRequest[] rqs = arrayOfrqs.getItem();
         RequestInfo info = getInfoFromRequest(rqs);
         no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSInformationRequest sr = (no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSInformationRequest)rqs[0];
         String requestId = sr.getRequestId();

         if ((info != null) && (info.scoreId != null) && (info.scoreId.length() > 0)) {
            DecisionMakerWebServices3Impl ws = getWs(protocol, hostname);
            no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSResponse response = ws.updateScoreDecisionById(request.getUser(), request.getPassWord(),
                  requestId, info.scoreId, info.finalResult, info.overruleComment);
            // Converting result back to XML
            return new WSRequestAndResponseClientWriter().responseToXML(response, rqs.length);
         }
         else {
            return "No scoreId set";
         }
      }
      else
         return null;
   }

   private static RequestInfo getInfoFromRequest(no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSSingleRequest[] rqs) {
      RequestInfo info = new RequestInfo();
      for (no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSSingleRequest rq : rqs) {
         if (rq instanceof no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSInformationRequest) {
            no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSInformationRequest irq = (no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSInformationRequest)rq;
            ArrayOfDmWSElement arrayOfaddInfos = irq.getAddInfo();

            if ((arrayOfaddInfos != null) && (arrayOfaddInfos.getItem() != null) && (arrayOfaddInfos.getItem().length > 0)) {
               DmWSElement[] addInfos = arrayOfaddInfos.getItem();
               for (no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSElement addInfo : addInfos) {
                  if ((addInfo != null) && (addInfo.getName() != null)) {
                     if (addInfo.getName().equals("id")) {
                        info.scoreId = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("finalResult")) {
                        info.finalResult = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("applicantFirstName")) {
                        info.firstName = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("applicantLastName")) {
                        info.lastName = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("created")) {
                        info.created = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("overrulecomment")) {
                        info.overruleComment = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("applicantId")) {
                        info.applicantId = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("applicantSsn")) {
                        info.applicantSsn = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("productId")) {
                        info.creditModelId = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("productName")) {
                        info.creditModelName = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("scoreGroupId")) {
                        info.scoreGroupId = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("scoreGroupName")) {
                        info.scoreGroupName = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("result")) {
                        info.scoreResult = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("scoreGroupId")) {
                        info.scoreGroupId = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("zone")) {
                        info.zone = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("overruledby")) {
                        info.overruledby = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("overruleddate")) {
                        info.overruleddate = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("score")) {
                        info.score = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("decision")) {
                        info.decision = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("points")) {
                        info.points = (String)addInfo.getVal();
                     }
                     else if (addInfo.getName().equals("createdBy")) {
                        info.createdBy = (String)addInfo.getVal();
                     }
                  }
               }
            }
         }
      }
      return info;
   }

   public static String scoreSingleFromXML(String protocol, String hostname, String requestXML) throws IOException, ServiceException,
         ExternalInterfacesXmlException {
      //Convert request XML to a WS request
      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSRequest request = new ScoringRequestXMLReader().fromXML(requestXML);

      ArrayOfDmWSSingleRequest arrayOfwsScoreRequestArray = request.getRequests();
      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSScoreRequest wsScoreRequest;

      if ((arrayOfwsScoreRequestArray != null) && (arrayOfwsScoreRequestArray.getItem() != null) && (arrayOfwsScoreRequestArray.getItem().length > 0)) {
         DmWSSingleRequest[] wsScoreRequestArray = arrayOfwsScoreRequestArray.getItem();
         wsScoreRequest = (no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSScoreRequest)wsScoreRequestArray[0];
      }
      else {
         wsScoreRequest = new no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSScoreRequest();
      }

      ArrayOfDmWSApplicant arrayOfdmWSApplicantArray = wsScoreRequest.getApplicants();
      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSApplicant applicant = null;

      if ((arrayOfdmWSApplicantArray != null) && (arrayOfdmWSApplicantArray.getItem() != null) && (arrayOfdmWSApplicantArray.getItem().length > 0)) {
         applicant = arrayOfdmWSApplicantArray.getItem()[0];
      }

      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSSingleResponse response = getWs(protocol, hostname).score(request.getUser(),
            request.getPassWord(), request.getUserRef(), wsScoreRequest.getRequestId(), applicant, wsScoreRequest.getProductName());
      // Converting result back to XML
      return new WSRequestAndResponseClientWriter().responseToXML(response);
   }

   public static String scoreGenericFromXML(String protocol, String hostname, String requestXML) throws IOException, ServiceException,
         ExternalInterfacesXmlException {
      // Convert request XML to a WS request
      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSRequest request = new ScoringRequestXMLReader().fromXML(requestXML);
      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSResponse response = getWs(protocol, hostname).scoreX(request);
      // Converting result back to XML
      return new WSRequestAndResponseClientWriter().responseToXML(response, 1);
   }

   public static String infoGenericFromXML(String protocol, String hostname, String requestXML) throws IOException, ServiceException,
         ExternalInterfacesXmlException {
      // Convert request XML to a WS request
      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSRequest request = new InformationRequestXMLReader().fromXML(requestXML);
      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSResponse response = getWs(protocol, hostname).getInformationX(request);
      // Converting result back to XML
      return new WSRequestAndResponseClientWriter().responseToXML(response, 1);
   }

   public static String infoSingleFromXML(String protocol, String hostname, String requestXML) throws IOException, ServiceException,
         ExternalInterfacesXmlException {
      // Convert request XML to a WS request
      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSRequest request = new InformationRequestXMLReader().fromXML(requestXML);

      DmWSSingleRequest[] srArray = request.getRequests().getItem();
      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSInformationRequest sr = null;

      if ((srArray != null) && (srArray.length > 0)) {
         sr = (no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSInformationRequest)srArray[0];
      }

      DmWSApplicant[] applicantArray = (sr == null) || (sr.getApplicants() == null) ? null : sr.getApplicants().getItem();
      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSApplicant applicant = null;

      if ((applicantArray != null) && (applicantArray.length > 0)) {
         applicant = applicantArray[0];
      }

      no.soliditet.wsclients.ws3.jaxrpc.generated.DmWSResponse response =
            getWs(protocol, hostname).getInformation(request.getUser(), request.getPassWord(), request.getUserRef(),
                  sr == null ? null : sr.getRequestId(), applicant, sr == null ? null : sr.getInformationHandler());

      // Converting result back to XML
      return new WSRequestAndResponseClientWriter().responseToXML(response, 1);
   }
}
