/*==========================================================================
 * Copyright (c) 2014, Bisnode Credit AS, All Rights Reserved.
 *
 * This software is the confidential and proprietary information
 * of Bisnode Credit AS ("Confidential Information"). You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Bisnode Credit AS.
 *
 * @since 01.jun.2006
=============================================================================*/
package no.soliditet.xmlclients.clientOne;

import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.security.Security;
import java.util.List;

import javax.xml.parsers.ParserConfigurationException;

import no.bisnode.credit.common.utils.StringUtil;
import no.soliditet.util.xml.XmlConstants;
import no.soliditet.util.xml.dom.XmlDOMDocument;
import no.soliditet.util.xml.dom.XmlDOMElement;

import org.apache.commons.lang.StringUtils;
import org.xml.sax.SAXException;

public class SoliditetDecision {

   protected PrintWriter out;
   protected URLConnection connection;
   protected boolean failed = false;
   private String urlstring;
   private String userName;
   private String password;

   public SoliditetDecision(String urlstring, String user, String password) {
      this.urlstring = urlstring;
      this.userName = user;
      this.password = password;
   }

   public String getScore(String orgNoOrSsn, String referanse, String kampanjeid, boolean personScore) throws IOException {
      if ((urlstring.contains("https")) || (urlstring.contains("HTTPS"))) {
         // Then we must use SSL
         prepareForSSL();
      }

      URL url = new URL(urlstring + "/scoring/score");
      try {
         out = openUrlConnection(url);
         System.out.println("connection opened ok");

      }
      catch (IOException exception) {
         // Too bad, something went wrong, probably is the server not running or the sertificate is wrong
         System.out.println("Error: " + exception);
         if (out == null) {
            failed = true;
         }
         else {
            System.out.println("Ignoring the error ...");
         }
         // Go on anyway
      }

      // Build the request
      XmlDOMDocument reqDoc = XmlDOMDocument.createDocumentWithRoot("dmRequest", XmlConstants.DEFAULT_ENCODING);
      XmlDOMElement rootElement = reqDoc.getRootElement();
      rootElement.addAttribute("version", "2.0");
      rootElement.addElement("user").setElementValue(userName);
      rootElement.addElement("password").setElementValue(password);
      XmlDOMElement requestElement = rootElement.addElement("request");
      requestElement.addAttribute("id", referanse);
      XmlDOMElement applicantElement = requestElement.addElement("applicant").addElement("APPLICANT");
      applicantElement.addElement("ssn").setElementValue(orgNoOrSsn);

      if (personScore) {
         applicantElement.addAttribute("type", "P");
         requestElement.addElement("productName").setElementValue("Standard Person");
      }
      else {
         applicantElement.addAttribute("type", "F");
         requestElement.addElement("productName").setElementValue("Standard Company");
      }

      if ((kampanjeid != null) && (kampanjeid.trim().length() > 0)) {
         requestElement.addElement("scoreGroupName").setElementValue(kampanjeid);
      }

      String request = reqDoc.getXmlWithHeader();
      // Run the request
      String sresult = doRequest(request);
      System.out.println("Result is " + sresult.length() + " characters long. ");

      return sresult;
   }

   /**
    * Use this for persons if you do not have the ssn
    * 
    * @param firstName
    * @param lastName
    * @param referanse
    * @param infoHandlerName
    * @return
    * @throws IOException
    */
   public String getInformation(String firstName, String lastName, String referanse, String infoHandlerName) throws IOException {
      return getInformation(null, firstName, lastName, referanse, true, infoHandlerName);
   }

   /**
    * Use this for companies if you don't have the orgno
    * 
    * @param name
    * @param referanse
    * @param infoHandlerName
    * @return
    * @throws IOException
    */
   public String getInformation(String name, String referanse, String infoHandlerName) throws IOException {
      return getInformation(null, null, name, referanse, false, infoHandlerName);
   }

   /**
    * Use this if you have the ssn or orgno
    * 
    * @param orgNoOrSsn
    * @param referanse
    * @param personInfo
    * @param infoHandlerName
    * @return
    * @throws IOException
    */
   public String getInformation(String orgNoOrSsn, String referanse, boolean personInfo, String infoHandlerName) throws IOException {
      return getInformation(orgNoOrSsn, null, null, referanse, personInfo, infoHandlerName);
   }

   public String getInformation(String orgNoOrSsn, String firstName, String lastName, String referanse, boolean personInfo, String infoHandlerName)
         throws IOException {
      if (StringUtils.containsIgnoreCase(urlstring, "https")) {
         // Then we must use SSL
         prepareForSSL();
      }

      URL url = new URL(urlstring + "/information/get");

      try {
         out = openUrlConnection(url);
         System.out.println("connection opened ok");

      }
      catch (IOException exception) {
         // Too bad, something went wrong, probably is the server not running or the sertificate is wrong
         System.out.println("Error: " + exception);
         if (out == null) {
            failed = true;
         }
         else {
            System.out.println("Ignoring the error ...");
         }
         // Go on anyway
      }

      XmlDOMDocument reqDoc = XmlDOMDocument.createDocumentWithRoot("dmRequest", XmlConstants.DEFAULT_ENCODING);
      XmlDOMElement rootElement = reqDoc.getRootElement();
      rootElement.addAttribute("version", "2.0");
      rootElement.addElement("user").setElementValue(userName);
      rootElement.addElement("password").setElementValue(password);
      XmlDOMElement requestElement = rootElement.addElement("request");
      requestElement.addAttribute("id", referanse);
      XmlDOMElement applicantElement = requestElement.addElement("applicant").addElement("APPLICANT");
      applicantElement.addElement("ssn").setElementValue(orgNoOrSsn);

      if (personInfo) {
         applicantElement.addAttribute("type", "P");
         applicantElement.addElement("firstName").setElementValue(firstName);
         applicantElement.addElement("lastName").setElementValue(lastName);
      }
      else {
         applicantElement.addAttribute("type", "F");
         applicantElement.addElement("name").setElementValue(lastName);
      }

      requestElement.addElement("informationHandler").setElementValue(infoHandlerName);

      String request = reqDoc.getXmlWithHeader();

      String sresult = doRequest(request);
      System.out.println("Result is " + sresult.length() + " characters long. ");

      return sresult;
   }

   private PrintWriter openUrlConnection(URL url) throws IOException {
      // Open the connection:
      System.out.println("opens connection to " + url);

      connection = url.openConnection();
      // and make it ready for output:
      connection.setDoOutput(true);

      // get a stream to send the data thru, and wrap it into a PrinterWriter:
      return new PrintWriter(connection.getOutputStream());
   }

   private String doRequest(String reqstring) {
      StringBuffer response = new StringBuffer(1000);
      if (!failed) {
         System.out.println("ready to send...\"" + reqstring + "\"");

         // Get the request, encode it .....
         String request = "request=" + URLencode(reqstring) + "\n";
         //... and send it:
         if ((out == null) || (failed)) {
            System.out.println("Failed to open connection!!");
         }
         else {
            out.print(request);
         }
      }

      // close the output stream
      if (out != null) {
         out.close();
      }

      if (!failed) {
         System.out.println("The request is sent!");

         // Now we are waiting for some answer ....

         try {
            // ... in which we need to read ...
            BufferedReader reader;
            try {
               //System.out.println("here4");
               System.out.println("Waiting for the response");
               // get an inputstrem and wrap into a BufferedReader:
               reader = new BufferedReader(new InputStreamReader(connection.getInputStream()));

               while (!reader.ready()) {
                  try {
                     Thread.sleep(100);
                  }
                  catch (InterruptedException e) {
                     // ignore exception
                  }
               }
               System.out.println("Response got! Reading ...");
            }
            catch (Exception exception) {
               InputStream err = ((HttpURLConnection)connection).getErrorStream();
               if (err == null) {
                  throw exception;
               }
               reader = new BufferedReader(new InputStreamReader(err));
            }

            // Save result in StringBuffer
            response = new StringBuffer(1000);
            String line;

            // read one and one line and add the newline char:
            while ((line = reader.readLine()) != null) {
               response.append(line).append("\n");
            }
            System.out.println("Response read!");

            // close the inputstream
            reader.close();
         }
         catch (Exception exception) {
            System.out.println("Error: " + exception);
         }
         finally {
            if (out != null) {
               out.close();
            }
         }
      }
      // and return the response:
      return response.toString();
   }

   public String URLencode(String s) {
      try {
         return URLEncoder.encode(s, "iso-8859-1");
      }
      catch (UnsupportedEncodingException e) {
         System.out.println("ENCODING EXCEPTION for string \"" + s + "\"" + e.getMessage());
         return s;
      }
   }

   private void prepareForSSL() {
      System.out.println("Running SSL here!");
      // set keystoretype:
      System.setProperty("javax.net.ssl.keyStoreType", "jks");
      // create new protocol handler:
      System.setProperty("java.protocol.handler.pkgs", "com.sun.net.ssl.internal.www.protocol");
      // add protocol handler:
      Security.addProvider(new com.sun.net.ssl.internal.ssl.Provider());
   }

   private static void getAndPrintInformation(SoliditetDecision ld, PrintStream stream, boolean printRawXml) {
      // its a company if false
      boolean personRequest = false;
      try {
         // call Soliditet Decision to get some info
         String response = ld.getInformation("868234202", "min ref", personRequest, "alpha:alle_web_co:xml:server");

         // now, let's see what we got.
         XmlDOMDocument rdoc = new XmlDOMDocument(response, XmlConstants.DEFAULT_ENCODING);
         // unless error, print info
         if (!printError(rdoc, stream, printRawXml)) {
            if (printRawXml) {
               // true means print xml as it comes from alpha
               printRawXml(stream, rdoc, true);
            }
            else {
               stream.println("Min referanse: " + rdoc.getXPathValue("/dmResponse/response/@id"));
               printAddressInfo(rdoc, personRequest, stream);

               // Nkkeltall
               printKeyFigures(rdoc, stream);

               printPaymentIndex(rdoc, stream);

               printDecisionScore(rdoc, stream);

               // add more
            }
         }
      }
      catch (IOException | ParserConfigurationException | SAXException e) {
         e.printStackTrace();
      }
   }

   private static void printRawXml(PrintStream stream, XmlDOMDocument rdoc, boolean alphaCompatible) {
      XmlDOMElement root = rdoc.getRootElement();
      if (alphaCompatible) {
         // then lets pick the <alpha> node
         XmlDOMElement alphaTag = root.getElementChildByName("alpha", 5);
         if (alphaTag != null) {
            root = alphaTag;
         }
      }
      stream.println(root.getXmlWithoutHeader());
   }

   private static void printDecisionScore(XmlDOMDocument rdoc, PrintStream stream) {
      String prefix = "/dmResponse/response/info/exti/alpha/cr/Rating3/";
      stream.println("Kredittsone: " + rdoc.getXPathValue(prefix + "CreditZone/@value"));
      stream.println("Vurdert dato: " + rdoc.getXPathValue(prefix + "Date/@value"));
      int noOfPolicyRules = convertToInt(rdoc.getXPathValue("count(" + prefix + "Comment/Comment)"));
      if (noOfPolicyRules > 0) {
         stream.println("Antall policyregler: " + noOfPolicyRules + ":");
      }
   }

   private static void printPaymentIndex(XmlDOMDocument rdoc, PrintStream stream) {
      String prefix = "/dmResponse/response/info/exti/alpha/cr/Rating2/";
      stream.println("Payment Index: " + rdoc.getXPathValue(prefix + "Rating/@value"));
      stream.println("Vurdert dato: " + rdoc.getXPathValue(prefix + "Date/@value"));
   }

   @SuppressWarnings({ "ForLoopReplaceableByForEach" })
   private static void printKeyFigures(XmlDOMDocument rdoc, PrintStream stream) {
      String prefix = "/dmResponse/response/info/exti/alpha/nt/";
      List keyFigureNodes = rdoc.getXPathList(prefix + "KeyFigures");
      for (int i = 0; i < keyFigureNodes.size(); i++) {
         XmlDOMElement keyFigure = XmlDOMElement.getAsElement(keyFigureNodes.get(i));
         if (keyFigure != null) {
            stream.println("Soliditet: " + keyFigure.getAttributeValueFromChild("Solidity", "value"));
            stream.println("Likviditetsgrad 1: " + keyFigure.getAttributeValueFromChild("LiquidityGrade1", "value"));
            // etc...
         }
      }
      stream.println("Payment Index: " + rdoc.getXPathValue(prefix + "Rating/@value"));
      stream.println("Vurdert dato: " + rdoc.getXPathValue(prefix + "Date/@value"));
   }

   private static boolean printError(XmlDOMDocument rdoc, PrintStream stream, boolean printRawXml) {
      if ((convertToInt(rdoc.getXPathValue("count(/dmResponse/error)")) > 0)
            || (convertToInt(rdoc.getXPathValue("count(/dmResponse/response/error)")) > 0)) {
         if (printRawXml) {
            printRawXml(stream, rdoc, false);
         }
         else {
            stream.println("Feilkode:" + rdoc.getXPathValue("//error/errorCode"));
            stream.println("Feilmelding:" + rdoc.getXPathValue("//error/errorMessage"));
            String whatToDo = rdoc.getXPathValue("//error/errorMeasure");
            if ((whatToDo != null) && (whatToDo.trim().length() > 0)) {
               stream.println("Hva skal gjres: " + whatToDo);
            }
         }
         return true;
      }
      else {
         // no error
         return false;
      }
   }

   private static void printAddressInfo(XmlDOMDocument rdoc, boolean personRequest, PrintStream stream) {
      String addressPrefix;
      if (personRequest) {
         addressPrefix = "/dmResponse/response/info/exti/alpha/ho/CompanyInfo";
      }
      else {
         addressPrefix = "/dmResponse/response/info/exti/alpha/ho/PersonInfo";
      }

      stream.println("Navn: " + rdoc.getXPathValue(addressPrefix + "/Name/@value"));
      stream.println("Adresse: " + rdoc.getXPathValue(addressPrefix + "/Address/@value"));
      stream.println("Postnr: " + rdoc.getXPathValue(addressPrefix + "/PostalNbr/@value"));
      stream.println("Sted: " + rdoc.getXPathValue(addressPrefix + "/PostalText/@value"));
   }

   /** @noinspection ConstantConditions, ForLoopReplaceableByForEach */
   private static void getAndPrintScoreResult(SoliditetDecision ld, PrintStream stream, boolean printRawXml) {
      // its a company if false
      boolean personRequest = false;
      try {
         // get the score
         String response = ld.getScore("868234202", "min ref", null, personRequest);
         XmlDOMDocument rdoc = new XmlDOMDocument(response, XmlConstants.DEFAULT_ENCODING);
         if (printError(rdoc, stream, printRawXml)) {
            return;
         }

         // no error, continue
         if (printRawXml) {
            printRawXml(stream, rdoc, false);
         }
         else {
            stream.println("Min referanse: " + rdoc.getXPathValue("/dmResponse/response/@id"));
            if (personRequest) {
               stream.println("Firmanavn: "
                     + rdoc.getXPathValue("/dmResponse/response/info/internal/dm/scoreresult/decision/SCORING/TOTALSCORE_ELEMENT/ELEMENT[@name='Navn']/@value"));
            }
            else {
               stream.println("Navn: "
                     + rdoc.getXPathValue("/dmResponse/response/info/internal/dm/scoreresult/decision/SCORING/TOTALSCORE_ELEMENT/ELEMENT[@name='Fornavn']/@value")
                     + ' '
                     + rdoc.getXPathValue("/dmResponse/response/info/internal/dm/scoreresult/decision/SCORING/TOTALSCORE_ELEMENT/ELEMENT[@name='Etternavnnavn']/@value"));
            }
            stream.println("Adresse: "
                  + rdoc.getXPathValue("/dmResponse/response/info/internal/dm/scoreresult/decision/SCORING/TOTALSCORE_ELEMENT/ELEMENT[@name='Adresse']/@value"));
            stream.println("Postnr: "
                  + rdoc.getXPathValue("/dmResponse/response/info/internal/dm/scoreresult/decision/SCORING/TOTALSCORE_ELEMENT/ELEMENT[@name='Postnr']/@value"));
            stream.println("Sted: "
                  + rdoc.getXPathValue("/dmResponse/response/info/internal/dm/scoreresult/decision/SCORING/TOTALSCORE_ELEMENT/ELEMENT[@name='Sted']/@value"));

            String decision = rdoc.getXPathValue("/dmResponse/response/info/internal/dm/scoreresult/result");
            switch (decision) {
               case "n":
                  decision = "Avslag";
                  break;
               case "c":
                  decision = "Kontroll";
                  break;
               case "y":
                  decision = "OK/Godkjent";
                  break;
            }
            stream.println("Beslutning: " + decision);

            int izone = convertToInt(rdoc.getXPathValue("/dmResponse/response/info/internal/dm/scoreresult/zone"));
            int modelNo = convertToInt(rdoc.getXPathValue("/dmResponse/response/info/internal/dm/scoreresult/decision/SCORING/ZONE_INFO/@modelNo"));

            String zoneText;
            if (modelNo == 2) {
               // Decision Payment Index
               // Payment Index er en scoringlsning som rangerer foretakene etter sannsynlighet for mislighold.
               switch (izone) {
                  case 1:
                     zoneText = "Rd sone";
                     break;
                  case 2:
                     zoneText = "Gul sone";
                     break;
                  case 3:
                     zoneText = "Grnn sone";
                     break;
                  default:
                     zoneText = "Payment Index ikke beregnet";
               }
            }
            else {
               // presumeably model 4 or 6, i.e. Decision SCore
               switch (izone) {
                  case 1:
                     zoneText = "Kreditt kun mot sikkerhet";
                     break;
                  case 2:
                     zoneText = "Kreditt mot sikkerhet anbefales";
                     break;
                  case 3:
                     zoneText = "Kredittverdig";
                     break;
                  case 4:
                     zoneText = "Hy kredittverdighet";
                     break;
                  case 5:
                     zoneText = "Hyeste kredittverdighet";
                     break;
                  default:
                     zoneText = "Foretaket er ikke ratet";
                     break;
               }
            }

            stream.println("Kredittsone: " + izone + ": " + zoneText);

            // Are there any policy rules occured?
            List list = rdoc.getXPathList("/dmResponse/response/info/internal/dm/scoreresult/decision/SCORING/POLICY_ELEMENT");
            if ((list != null) && (list.size() > 0)) {
               stream.println("Flgende policyregler foreligger:");
               for (int i = 0; i < list.size(); i++) {
                  XmlDOMElement po = XmlDOMElement.getAsElement(list.get(i));
                  if (po != null) {
                     // name of policyrule, and info (inside the "action" attribute)
                     stream.println(po.getAttributeValue("name") + ": " + po.getAttributeValue("action"));
                  }
               }
            }
         }
      }
      catch (IOException | ParserConfigurationException | SAXException e) {
         e.printStackTrace();
      }
   }

   private static int convertToInt(String numberAsString) {
      // 0 is default
      int inumber = 0;

      try {
         // expect to get floating-point value from xpath
         inumber = (int)Double.parseDouble(numberAsString);
      }
      catch (NumberFormatException e) {
         try {
            double doubleValue = StringUtil.convertStringToDouble(numberAsString);
            // skipping decimals
            inumber = (int)doubleValue;
         }
         catch (NumberFormatException e1) {
            // ignore exception
         }
      }

      return inumber;
   }

   public static void main(String[] args) {
      String hostName = null;
      if ((args != null) && (args.length > 0) && (args[0].trim().length() > 0)) {
         hostName = args[0].trim();
      }
      if ((hostName == null) || (hostName.trim().length() == 0)) {
         hostName = "https://test.soliditetd.no";
      }

      SoliditetDecision ld = new SoliditetDecision(hostName, "system", "system");

      boolean printRawXml = true;
      PrintStream infoStream = System.out;
      PrintStream scoreStream = System.out;
      boolean outputToFile = true;

      if (outputToFile) {
         try {
            infoStream = new PrintStream(new BufferedOutputStream(new FileOutputStream("resultInfo.xml")));
            scoreStream = new PrintStream(new BufferedOutputStream(new FileOutputStream("resultScore.xml")));
         }
         catch (FileNotFoundException e) {
            // print file open error
            e.printStackTrace();
         }
      }
      getAndPrintInformation(ld, infoStream, printRawXml);

      getAndPrintScoreResult(ld, scoreStream, printRawXml);

      if (outputToFile) {
         infoStream.flush();
         infoStream.close();
         scoreStream.flush();
         scoreStream.close();
      }
   }
}
